<!-- katex links / this needs to be loaded before fonts.html -->

<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/katex.min.css" crossorigin="anonymous">
<script async src="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/katex.min.js" crossorigin="anonymous"></script>

<!-- Preload the theme fonts -->

<link rel="preload" href="{{ pathto('_static/fonts/FreightSans/freight-sans-book.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/FreightSans/freight-sans-medium.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/IBMPlexMono/IBMPlexMono-Medium.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/FreightSans/freight-sans-bold.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/FreightSans/freight-sans-medium-italic.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/IBMPlexMono/IBMPlexMono-SemiBold.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="{{ pathto('_static/fonts/IBMPlexMono/IBMPlexMono-Regular.woff2', 1) }}" as="font" type="font/woff2" crossorigin="anonymous">

<!-- Preload the katex fonts -->

<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Math-Italic.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Main-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Main-Bold.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Size1-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Size4-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Size2-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Size3-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
<link rel="preload" href="https://cdn.jsdelivr.net/npm/katex@0.13.20/dist/fonts/KaTeX_Caligraphic-Regular.woff2" as="font" type="font/woff2" crossorigin="anonymous">
