{# TEMPLATE VAR SETTINGS #}
{%- set url_root = pathto('', 1) %}
{%- if url_root == '#' %}{% set url_root = '' %}{% endif %}
{%- if not embedded and docstitle %}
  {%- set titlesuffix = " &mdash; "|safe + docstitle|e %}
{%- else %}
  {%- set titlesuffix = "" %}
{%- endif %}
{%- set lang_attr = 'en' if language == None else (language | replace('_', '-')) %}
{% import 'theme_variables.jinja' as theme_variables %}
{% set show_netlify_badge = True %}

<!DOCTYPE html>
<!--[if IE 8]><html class="no-js lt-ie9" lang="{{ lang_attr }}" > <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="{{ lang_attr }}" > <!--<![endif]-->
<head>
  <meta charset="utf-8">
  <meta name="author" content="PyTorch-Ignite Contributors">
  <meta name="creator" content="PyTorch-Ignite Contributors">
  <meta name="publisher" content="PyTorch-Ignite Contributors">
  <meta name="generator" content="Sphinx {{ sphinx_version }}">
  <meta name="description" content="High-level library to help with training and evaluating neural networks in PyTorch flexibly and transparently.">
  <meta name="keywords" content="pytorch, pytorch ignite, ignite, engine, events, handlers, metrics">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <meta name="theme-color" content="#ee4c2c">

  <meta name="twitter:card" content="summary_large_image">
  <meta name="twitter:site" content="@pytorch_ignite">
  <meta name="twitter:creator" content="@pytorch_ignite">
  <meta name="twitter:description" content="High-level library to help with training and evaluating neural networks in PyTorch flexibly and transparently.">
  <meta name="twitter:title" content="{{ title|striptags|e }}{{ titlesuffix }}">
  <meta name="twitter:image" content="https://raw.githubusercontent.com/pytorch/ignite/master/assets/logo/ignite_logo.png">
  <meta name="twitter:image:alt" content="PyTorch-Ignite logo">

  <meta property="og:title" content="{{ title|striptags|e }}{{ titlesuffix }}">
  <meta property="og:type" content="website">
  <meta property="og:url" content="https://pytorch.org/ignite/">
  <meta property="og:site_name" content="PyTorch-Ignite">
  <meta property="og:image" content="https://raw.githubusercontent.com/pytorch/ignite/master/assets/logo/ignite_logo.png">
  <meta property="og:image:alt" content="PyTorch-Ignite logo">
  <meta property="og:description" content="High-level library to help with training and evaluating neural networks in PyTorch flexibly and transparently.">

  <link rel="preconnect" href="https://BH4D9OD16A-dsn.algolia.net" crossorigin />

  {% block htmltitle %}
  <title>{{ title|striptags|e }}{{ titlesuffix }}</title>
  {% endblock %}

  {# FAVICON #}
  {% if favicon %}
    <link rel="shortcut icon" type="image/svg+xml" href="{{ pathto('_static/' + favicon, 1) }}"/>
  {% endif %}
  {# CANONICAL URL #}
  {% if theme_canonical_url %}
    <link rel="canonical" href="{{ theme_canonical_url }}{{ pagename }}.html"/>
  {% endif %}

  {# CSS #}

  {# OPENSEARCH #}
  {% if not embedded %}
    {% if use_opensearch %}
      <link rel="search" type="application/opensearchdescription+xml"
            title="{% trans docstitle=docstitle|e %}Search within {{ docstitle }}{% endtrans %}"
            href="{{ pathto('_static/opensearch.xml', 1) }}"/>
    {% endif %}

  {% endif %}

  <link rel="preload" href="{{ pathto('_static/' + style, 1) }}" as="style" />
  <link rel="stylesheet" href="{{ pathto('_static/' + style, 1) }}" type="text/css" />
  <!-- <link rel="stylesheet" href="{{ pathto('_static/pygments.css', 1) }}" type="text/css" /> -->
  {%- for css in css_files %}
    {%- if css|attr("rel") %}
  <link rel="preload" href="{{ pathto(css.filename, 1) }}" as="style" />
  <link rel="{{ css.rel }}" href="{{ pathto(css.filename, 1) }}" type="text/css"{% if css.title is not none %} title="{{ css.title }}"{% endif %} />
    {%- else %}
  <link rel="preload" href="{{ pathto(css, 1) }}" as="style" />
  <link rel="stylesheet" href="{{ pathto(css, 1) }}" type="text/css" />
    {%- endif %}
  {%- endfor %}
  {%- for cssfile in extra_css_files %}
    <link rel="preload" href="{{ pathto(cssfile, 1) }}" as="style" />
    <link rel="stylesheet" href="{{ pathto(cssfile, 1) }}" type="text/css" />
  {%- endfor %}

  {%- block linktags %}
    {%- if hasdoc('about') %}
    <link rel="author" title="{{ _('About these documents') }}" href="{{ pathto('about') }}" />
    {%- endif %}
    {%- if hasdoc('genindex') %}
    <link rel="index" title="{{ _('Index') }}" href="{{ pathto('genindex') }}" />
    {%- endif %}
    {%- if hasdoc('search') %}
    <link rel="search" title="{{ _('Search') }}" href="{{ pathto('search') }}" />
    {%- endif %}
    {%- if hasdoc('copyright') %}
    <link rel="copyright" title="{{ _('Copyright') }}" href="{{ pathto('copyright') }}" />
    {%- endif %}
    {%- if next %}
    <link rel="next" title="{{ next.title|striptags|e }}" href="{{ next.link|e }}" />
    {%- endif %}
    {%- if prev %}
    <link rel="prev" title="{{ prev.title|striptags|e }}" href="{{ prev.link|e }}" />
    {%- endif %}
  {%- endblock %}
  {%- block extrahead %} {% endblock %}

  {% include "fonts.html" %}

  <!-- Global site tag (gtag.js) - Google Analytics -->
  <script async src="https://www.googletagmanager.com/gtag/js?id=G-5FELLLFHCP"></script>
  <script>
    window.dataLayer = window.dataLayer || [];
    function gtag(){dataLayer.push(arguments);}
    gtag('js', new Date());

    gtag('config', 'G-5FELLLFHCP');
  </script>
</head>

<div class="container-fluid header-holder tutorials-header" id="header-holder">
  <div class="container">

    <a class="header-logo" href="/ignite" aria-label="PyTorch-Ignite"></a>

    <div class="header-container">

      <div class="main-menu">
        <ul>
          <li>
            <a href="{{ theme_variables.external_urls['home'] }}/{{ theme_variables.external_urls['quickstart'] }}">Quickstart</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['home'] }}/{{ theme_variables.external_urls['concepts'] }}">Concepts</a>
          </li>

          <!-- <li>
            <a href="{{ theme_variables.external_urls['home'] }}/{{ theme_variables.external_urls['examples'] }}">Examples</a>
          </li> -->

          <li>
            <a href="{{ theme_variables.external_urls['home'] }}/{{ theme_variables.external_urls['faq'] }}">FAQ</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['github'] }}" target="_blank" rel="noopener noreferrer">GitHub</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['home'] }}/{{ theme_variables.external_urls['about_us'] }}">About us</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['home'] }}">⊳ pytorch-ignite.ai</a>
          </li>


        </ul>
      </div>

      <!-- <a class="main-menu-open-button" href="#" data-behavior="open-mobile-menu"></a> -->
    </div>

  </div>
</div>


<body class="pytorch-body">

  {% block extrabody %} {% endblock %}

    {# SIDE NAV, TOGGLES ON MOBILE #}

    <div class="table-of-contents-link-wrapper">
      <span>Table of Contents</span>
      <div class="toggle-table-of-contents" data-behavior="toggle-table-of-contents"></div>
    </div>

    <nav data-toggle="wy-nav-shift" class="pytorch-left-menu" id="pytorch-left-menu">
      <div class="pytorch-side-scroll">
        <div class="pytorch-menu pytorch-menu-vertical" data-spy="affix" role="navigation" aria-label="main navigation">
          <div class="pytorch-left-menu-search">
            {% block sidebartitle %}

            {% if theme_display_version %}
              {%- set nav_version = version %}
              {% if READTHEDOCS and current_version %}
                {%- set nav_version = current_version %}
              {% endif %}
              {% if nav_version %}
                <div class="version">
                  {{ nav_version }}
                </div>
              {% endif %}
            {% endif %}

            <div id="docsearch"></div>

            {% endblock %}
          </div>

          {% block menu %}
            {#
              The singlehtml builder doesn't handle this toctree call when the
              toctree is empty. Skip building this for now.
            #}
            {% if 'singlehtml' not in builder %}
              {% set global_toc = toctree(maxdepth=1,
                                          collapse=theme_collapse_navigation|tobool,
                                          includehidden=theme_includehidden|tobool,
                                          titles_only=theme_titles_only|tobool) %}
            {% endif %}
            {% if global_toc %}
              {{ global_toc }}
            {% else %}
              <!-- Local TOC -->
              <div class="local-toc">{{ toc }}</div>
            {% endif %}
          {% endblock %}
        </div>
      </div>

      {% include "versions.html" %}

    </nav>

    <div class="pytorch-container">
      <div class="pytorch-page-level-bar" id="pytorch-page-level-bar">
        <div class="pytorch-breadcrumbs-wrapper">
          {% include "breadcrumbs.html" %}
        </div>

        <div class="pytorch-shortcuts-wrapper" id="pytorch-shortcuts-wrapper">
          Shortcuts
        </div>
      </div>

      <section data-toggle="wy-nav-shift" id="pytorch-content-wrap" class="pytorch-content-wrap">
        <div class="pytorch-content-left">

        {% if theme_pytorch_project == 'tutorials' %}

          <div class="pytorch-call-to-action-links">
            <div id="tutorial-type">{{ pagename }}</div>

            <div id="google-colab-link">
              <img class="call-to-action-img" src="{{ pathto('_static/images/pytorch-colab.svg', 1) }}"/>
              <div class="call-to-action-desktop-view">Run in Google Colab</div>
              <div class="call-to-action-mobile-view">Colab</div>
            </div>
            <div id="download-notebook-link">
              <img class="call-to-action-notebook-img" src="{{ pathto('_static/images/pytorch-download.svg', 1) }}"/>
              <div class="call-to-action-desktop-view">Download Notebook</div>
              <div class="call-to-action-mobile-view">Notebook</div>
            </div>
            <div id="github-view-link">
              <img class="call-to-action-img" src="{{ pathto('_static/images/pytorch-github.svg', 1) }}"/>
              <div class="call-to-action-desktop-view">View on GitHub</div>
              <div class="call-to-action-mobile-view">GitHub</div>
            </div>
          </div>

        {% endif %}

        {%- block content %}
          {% if theme_style_external_links|tobool %}
          <div class="rst-content style-external-links">
          {% else %}
          <div class="rst-content">
          {% endif %}
            <div role="main" class="main-content" itemscope="itemscope" itemtype="http://schema.org/Article">
            {%- block document %}
             <article itemprop="articleBody" id="pytorch-article" class="pytorch-article">
              {% block body %}{% endblock %}
             </article>
             {% if self.comments()|trim %}
             <div class="articleComments">
              {% block comments %}{% endblock %}
             </div>
             {% endif%}
            </div>
            {%- endblock %}
            {% include "footer.html" %}
          </div>
        {%- endblock %}
        </div>

        <div class="pytorch-content-right" id="pytorch-content-right">
          <div class="pytorch-right-menu" id="pytorch-right-menu">
            <div class="pytorch-side-scroll" id="pytorch-side-scroll-right">
              {{ toc }}
            </div>
          </div>
        </div>
      </section>
    </div>

  {% if not embedded %}

     {% if sphinx_version >= "1.8.0" %}
       <script type="text/javascript" id="documentation_options" data-url_root="{{ pathto('', 1) }}" src="{{ pathto('_static/documentation_options.js', 1) }}"></script>
       {%- for scriptfile in script_files %}
         {{ js_tag(scriptfile) }}
       {%- endfor %}
     {% else %}
       <script type="text/javascript">
           var DOCUMENTATION_OPTIONS = {
               URL_ROOT:'{{ url_root }}',
               VERSION:'{{ release|e }}',
               LANGUAGE:'{{ language }}',
               COLLAPSE_INDEX:false,
               FILE_SUFFIX:'{{ '' if no_search_suffix else file_suffix }}',
               HAS_SOURCE:  {{ has_source|lower }},
               SOURCELINK_SUFFIX: '{{ sourcelink_suffix }}'
           };
       </script>
       {%- for scriptfile in script_files %}
         <script type="text/javascript" src="{{ pathto(scriptfile, 1) }}"></script>
       {%- endfor %}
     {% endif %}

  {% endif %}

  <script type="text/javascript" src="{{ pathto('_static/js/vendor/popper.min.js', 1) }}"></script>
  <script type="text/javascript" src="{{ pathto('_static/js/vendor/bootstrap.min.js', 1) }}"></script>
  <script type="text/javascript" src="{{ pathto('_static/js/theme.js', 1) }}"></script>

  <script type="text/javascript">
      jQuery(function () {
          SphinxRtdTheme.Navigation.enable({{ 'true' if theme_sticky_navigation|tobool else 'false' }});
      });
  </script>

  {%- block footer %} {% endblock %}

  <!-- Begin Footer -->

  <!-- commented out for Ignite -->

  <div class="container-fluid docs-tutorials-resources" id="docs-tutorials-resources">
    <!-- <div class="container">
      <div class="row">
        <div class="text-center col-md-4">
          <h2>Docs</h2>
          <p>Access comprehensive developer documentation for PyTorch</p>
          <a class="with-right-arrow" href="{{ theme_variables.external_urls['docs'] }}">View Docs</a>
        </div>

        <div class="text-center col-md-4">
          <h2>Tutorials</h2>
          <p>Get in-depth tutorials for beginners and advanced developers</p>
          <a class="with-right-arrow" href="{{ theme_variables.external_urls['tutorials'] }}">View Tutorials</a>
        </div>

        <div class="text-center col-md-4">
          <h2>Resources</h2>
          <p>Find development resources and get your questions answered</p>
          <a class="with-right-arrow" href="{{ theme_variables.external_urls['resources'] }}">View Resources</a>
        </div>
      </div>
    </div> -->
  </div>

  <!-- <footer class="site-footer">
    <div class="container footer-container">
      <div class="footer-logo-wrapper">
        <a href="{{ theme_variables.external_urls['home'] }}" class="footer-logo"></a>
      </div>

      <div class="footer-links-wrapper">
        <div class="footer-links-col">
          <ul>
            <li class="list-title"><a href="{{ theme_variables.external_urls['home'] }}">PyTorch</a></li>
            <li><a href="{{ theme_variables.external_urls['get_started'] }}">Get Started</a></li>
            <li><a href="{{ theme_variables.external_urls['features'] }}">Features</a></li>
            <li><a href="{{ theme_variables.external_urls['ecosystem'] }}">Ecosystem</a></li>
            <li><a href="{{ theme_variables.external_urls['blog'] }}">Blog</a></li>
            <li><a href="{{ theme_variables.external_urls['resources'] }}">Resources</a></li>
          </ul>
        </div>

        <div class="footer-links-col">
          <ul>
            <li class="list-title"><a href="{{ theme_variables.external_urls['support'] }}">Support</a></li>
            <li><a href="{{ theme_variables.external_urls['tutorials'] }}">Tutorials</a></li>
            <li><a href="{{ theme_variables.external_urls['docs'] }}">Docs</a></li>
            <li><a href="{{ theme_variables.external_urls['discuss'] }}" target="_blank">Discuss</a></li>
            <li><a href="{{ theme_variables.external_urls['github_issues'] }}" target="_blank">Github Issues</a></li>
            <li><a href="{{ theme_variables.external_urls['slack'] }}" target="_blank">Slack</a></li>
            <li><a href="{{ theme_variables.external_urls['contributing'] }}" target="_blank">Contributing</a></li>
          </ul>
        </div>

        <div class="footer-links-col follow-us-col">
          <ul>
            <li class="list-title">Follow Us</li>
            <li>
              <div id="mc_embed_signup">
                <form
                  action="https://twitter.us14.list-manage.com/subscribe/post?u=75419c71fe0a935e53dfa4a3f&id=91d0dccd39"
                  method="post"
                  id="mc-embedded-subscribe-form"
                  name="mc-embedded-subscribe-form"
                  class="email-subscribe-form validate"
                  target="_blank"
                  novalidate>
                  <div id="mc_embed_signup_scroll" class="email-subscribe-form-fields-wrapper">
                    <div class="mc-field-group">
                      <label for="mce-EMAIL" style="display:none;">Email Address</label>
                      <input type="email" value="" name="EMAIL" class="required email" id="mce-EMAIL" placeholder="Email Address">
                    </div>

                    <div id="mce-responses" class="clear">
                      <div class="response" id="mce-error-response" style="display:none"></div>
                      <div class="response" id="mce-success-response" style="display:none"></div>
                    </div>    <!~~ real people should not fill this in and expect good things - do not remove this or risk form bot signups ~~>

                    <div style="position: absolute; left: -5000px;" aria-hidden="true"><input type="text" name="b_75419c71fe0a935e53dfa4a3f_91d0dccd39" tabindex="-1" value=""></div>

                    <div class="clear">
                      <input type="submit" value="" name="subscribe" id="mc-embedded-subscribe" class="button email-subscribe-button">
                    </div>
                  </div>
                </form>
              </div>

            </li>
          </ul>

          <div class="footer-social-icons">
            <a href="{{ theme_variables.external_urls['facebook'] }}" target="_blank" class="facebook"></a>
            <a href="{{ theme_variables.external_urls['twitter'] }}" target="_blank" class="twitter"></a>
          </div>
        </div>
      </div>
    </div>
  </footer> -->


  <!-- end of commented out for Ignite -->

  <!-- {% include "cookie_banner.html" %} -->

  <!-- End Footer -->

  <!-- Begin Mobile Menu -->
  <!--
  <div class="mobile-main-menu">
    <div class="container-fluid">
      <div class="container">
        <div class="mobile-main-menu-header-container">
          <a class="header-logo" href="{{ theme_variables.external_urls['home'] }}" aria-label="PyTorch"></a>
          <a class="main-menu-close-button" href="#" data-behavior="close-mobile-menu"></a>
        </div>
      </div>
    </div>

    <div class="mobile-main-menu-links-container">
      <div class="main-menu">
        <ul>
          <li>
            <a href="#">Get Started</a>
          </li>

          <li>
            <a href="#">Features</a>
          </li>

          <li>
            <a href="#">Ecosystem</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['blog'] }}">Blog</a>
          </li>

          <li {%- if theme_pytorch_project == 'tutorials' %} class="active"{%- endif %}>
            <a href="{{ theme_variables.external_urls['tutorials'] }}">Tutorials</a>
          </li>

          <li {%- if theme_pytorch_project == 'docs' %} class="active"{%- endif %}>
            <a href="{{ theme_variables.external_urls['docs'] }}">Docs</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['resources'] }}">Resources</a>
          </li>

          <li>
            <a href="{{ theme_variables.external_urls['github'] }}">Github</a>
          </li>
        </ul>
      </div>
    </div>
  </div>
  -->
  <!-- End Mobile Menu -->

  <script type="text/javascript" src="{{ pathto('_static/js/vendor/anchor.min.js', 1) }}"></script>

  <script type="text/javascript">
    var collapsedSections = []
    $(document).ready(function() {
      mobileMenu.bind();
      mobileTOC.bind();
      pytorchAnchors.bind();
      sideMenus.bind();
      scrollToAnchor.bind();
      highlightNavigation.bind();

      // Add class to links that have code blocks, since we cannot create links in code blocks
      $("article.pytorch-article a span.pre").each(function(e) {
        $(this).closest("a").addClass("has-code");
      });
    })
  </script>
  <script src="https://cdn.jsdelivr.net/npm/@docsearch/js@3"></script>
  <script type="text/javascript">
  let VERSION
  if ('{{ version }}'.startsWith('v')) {
    VERSION = '{{ version }}'
  } else {
    VERSION = 'master'
  }
  docsearch({
    container: '#docsearch',
    appId: '7EWYE1JCT3',
    apiKey: '841e93e60c16975ba1bd8c7c716eed82',
    indexName: 'pytorch-ignite',
    placeholder: 'Search PyTorch-Ignite docs',
    searchParameters: {
      facetFilters: [`version:${VERSION}`, 'tags:API-reference'],
    }
  });
  </script>
</body>
</html>
