{%- set external_urls = {
  'github': 'https://github.com/pytorch/ignite',
  'github_issues': 'https://github.com/pytorch/ignite/issues',
  'contributing': 'https://github.com/pytorch/ignite/blob/master/CONTRIBUTING.md',
  'docs': 'https://pytorch.org/ignite/index.html',

  'home': 'https://pytorch-ignite.ai',

  'concepts': 'concepts/',
  'quickstart': 'tutorials/beginner/01-getting-started/',
  'examples': 'tutorials/beginner/01-getting-started/#complete-code',
  'faq': 'how-to-guides/',
  'about_us': 'about/community/'
}
-%}
