article.pytorch-article table tr th:first-of-type,
article.pytorch-article table tr td:first-of-type {
  width: 17px;
}

div.container a.header-logo {
  height: 80px;
  width: 160px;
  background-image: url("../img/ignite_logo.svg");
  background-size: 160px;
}

/* versions block */

.rst-versions {
  position: fixed;
  bottom: 0;
  left: 0;
  width: inherit;
  color: #fcfcfc;
  font-family: FreightSans, Helvetica Neue, Helvetica, Arial, sans-serif;
  background: #d1d2d5;
  z-index: 400;
}

.rst-versions.shift-up {
  height: auto;
  max-height: 100%;
  overflow-y: scroll;
}

.rst-versions.shift-up .rst-other-versions {
  display: block;
}

.rst-versions .rst-other-versions {
  font-size: 90%;
  padding: 12px;
  color: #6c6c6d;
  background: #f2f3f7;
  display: none;
}

.rst-versions .rst-current-version {
  padding: 12px;
  background-color: #e2e3e6;
  display: block;
  text-align: right;
  font-size: 90%;
  cursor: pointer;
  color: #ee4c2c;
}

.rst-versions .rst-other-versions dd {
  display: inline-block;
  margin: 0;
}

.rst-versions .rst-other-versions dd a {
  display: inline-block;
  padding: 6px;
  color: #454545;
}

.rst-versions a {
  color: #2980b9;
  text-decoration: none;
}

.rst-versions .rst-current-version:after,
.rst-versions .rst-current-version:before {
  display: table;
  content: "";
}

.rst-versions .rst-current-version:after {
  clear: both;
  content: "";
  display: block;
}

.rst-versions .rst-current-version .fa-book,
.rst-versions .rst-current-version .icon-book {
  float: left;
}

.rst-versions .rst-current-version .fa {
  color: #454545;
}

.rst-versions .rst-current-version .fa-book,
.rst-versions .rst-current-version .icon-book {
  float: left;
}

.rst-other-versions dt {
  font-weight: normal;
}

.fa {
  font-family: inherit;
  display: inline-block;
  /* font: normal normal normal 14px/1 FontAweSome; */
  font-size: inherit;
  text-rendering: auto;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}

.fa-book:before, .icon-book: before {
  content: "\f02d";
}

.fa-caret-down:before {
  content: "▼";
  color: #ee4c2c;
}

.fa:before,
a .fa,
li .fa {
  text-decoration: inherit;
}

.fa:before {
  font-family: FontAwesome;
  font-style: normal;
  font-weight: 400;
  line-height: 1;
}

/* automatically generated toctree tables */

article.pytorch-article table.longtable.docutils.align-default colgroup {
  display: none;
}

article.pytorch-article
  table.longtable.docutils.align-default
  tbody
  td:first-child {
  width: 40%;
}

article.pytorch-article table.longtable.docutils.align-default tbody td {
  padding: 0.25rem;
}

article.pytorch-article table.longtable.docutils.align-default tbody td p {
  margin-block-start: 0.1em;
  margin-block-end: 0.1em;
}

article.pytorch-article div.highlight pre {
  white-space: pre;
}

/* versionadded, versionchanged, deprecated directive */

.versionmodified {
  font-style: italic;
  font-weight: bold;
}

/* Footer */

footer table {
  width: 100%;
  margin-bottom: 1rem;
}

/* remove gray footer */
.pytorch-body .docs-tutorials-resources {
  padding: 0 !important;
}

/* dim out pytorch header
algolia searchbox has z-index of 200
so make sure not to go above that value */
.header-holder {
  z-index: 100 !important;
}

:root {
  --docsearch-primary-color: #ee4c2c !important;
}

.DocSearch-Button {
  border-radius: 8px !important;
  width: 100% !important;
  height: 2.75rem !important;
  margin: 0 !important;
  background: var(--docsearch-searchbox-focus-background) !important;
}

.DocSearch-Button .DocSearch-Search-Icon {
  color: var(--docsearch-primary-color) !important;
}

.DocSearch-Button-Key {
  background: none !important;
  box-shadow: none !important;
  top: 0 !important;
  padding-bottom: 0 !important;
  color: inherit !important;
  font-size: 1rem !important;
  font-family: Arial !important;
}

@media (max-width: 750px) {
  .DocSearch-Button-Key,
  .DocSearch-Button-Placeholder {
    display: block !important;
  }
}
/* docsearch end */

article.pytorch-article p.rubric {
  font-weight: bold;
  font-size: 1.25rem;
  margin: 15px 0 10px 0;
}

.highlight pre {
  border: 1px solid rgba(0, 0, 0, 0.15);
  border-radius: 4px;
  line-height: 1.375rem;
}

/* Sphinx >=3.5: Fix class/method/fn names, because of spans */
article.pytorch-article .class dt > span span.pre,
article.pytorch-article .function dt > span span.pre {
  color: inherit;
  font-size: 87.5%;
  font-family: IBMPlexMono, SFMono-Regular, Menlo, Monaco, Consolas,
    "Liberation Mono", "Courier New", monospace;
}

/* Sphinx >=3.5: Fix annotations */
article.pytorch-article .class dt span.pre,
article.pytorch-article .function dt span.pre {
  color: inherit;
  padding: 2px 0.2px;
}
